"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2019-2019 by Zhong-Li Liu

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""

# get isovalue lines of Gibbs free energy from contour plot
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import griddata
import json
import sys


def generate_TGP(phase_type, dictionary):
    """
    Generate temperature, Gibbs free energy, and pressure data for a given phase.

    Parameters:
    phase_type (str): The phase type ('solid' or 'liquid').
    dictionary (dict): The dictionary containing data.

    Returns:
    tuple: Meshgrid arrays for temperature, Gibbs free energy, and interpolated pressure data.
    """
    # Input data
    grid_dat = dictionary[phase_type]['dat']
    dat = np.array(grid_dat)

    # Solid phase or Liquid phase
    X = dat[:, 3]  # Temperature
    Y = dat[:, 7]  # Gibbs free energy
    Z = dat[:, 2]  # Pressure

    # Interpolating data
    xi = np.linspace(X.min(), X.max(), 60)
    yi = np.linspace(Y.min(), Y.max(), 60)
    # zi = griddata((X, Y), Z, (xi[None, :], yi[:, None]), method='cubic', )
    zi = griddata((X, Y), Z, (xi[None, :], yi[:, None]), method='cubic', )
    xig, yig = np.meshgrid(xi, yi)

    return xig, yig, zi


def save_contour_lines(phase, level, dictionary):
    """
    Save contour lines for a given phase at a specified pressure level to a file.

    Parameters:
    phase (str): The phase type ('solid' or 'liquid').
    level (float): The pressure level to generate contour lines for.
    dictionary (dict): The dictionary containing data.
    """
    try:
        xig, yig, zi = generate_TGP(phase, dictionary)
        cset = plt.contour(xig, yig, zi, [level])
        level_value_str = f"{level * 10 :.3f}"
        file_name = f"{phase}_pressure_{level_value_str}.txt"

        path = cset.collections[0].get_paths()[0]
        vert = path.vertices

        with open(file_name, 'w') as f:
            for j in range(vert.shape[0]):
                # f.write(f"{vert[j, 0]:.6f} {vert[j, 1]:.6f}\n")
                f.write(f"{vert[j, 0]} {vert[j, 1]}\n")
    except Exception as e:
        print(f"An error occurred while saving contour lines for {phase}: {e}", file=sys.stderr)


def contour_line(p, n):
    """
    Generate and save contour lines of Gibbs free energy for both solid and liquid phases.

    Parameters:
    p (float): The pressure level.
    n (str): The filename prefix for the input JSON file.
    """
    try:
        with open(f'{n}.json', 'r') as t:
            dictionary = json.load(t)

        # Specify isovalue level of pressure
        level = float(p) * 1.0e-1  # GPa ==> Pa*1e+10

        # Save contour lines for both phases
        save_contour_lines('solid', level, dictionary)
        save_contour_lines('liquid', level, dictionary)

    except Exception as e:
        print(f"An error occurred: {e}", file=sys.stderr)




import numpy as np
from scipy.optimize import curve_fit, fsolve
import warnings

warnings.filterwarnings('ignore')


def func_US(S, a, b, c):
    U = c * np.exp(S / a) + b
    return U


def func_TS(S, a, c):
    T = c / a * np.exp(S / a)
    return T


def func_ST(T, a, c):
    S = a * np.log(a * T / c)
    return S


# BirchMurnaghan equation from PRB 70, 224107
def func_UV(V, E0, B0, BP, V0):
    eta = (V0 / V) ** (1. / 3.)
    U = E0 + 9. * B0 * V0 / 16. * ((eta ** 2 - 1) ** 3 * BP + (eta ** 2 - 1) ** 2 * (6 - 4 * eta ** 2))
    return U


def func_pV(V, B0, BP, V0):
    eta = (V0 / V) ** (1. / 3.)
    p = 1.5 * B0 * (eta ** 7 - eta ** 5) * (1 + 3. / 4. * (BP - 4) * (eta ** 2 - 1))
    return p


def func_Vp(p, B0, BP, V0, dat_abc):
    minV = min(dat_abc[:, 0])
    maxV = max(dat_abc[:, 0])

    def f(x):
        eta = (V0 / x) ** (1. / 3.)
        return 1.5 * B0 * (eta ** 7 - eta ** 5) * (1 + 3. / 4. * (BP - 4) * (eta ** 2 - 1)) - p

    # determine the guess volume
    n_V = 50
    V_list = np.linspace(minV, maxV, n_V)
    delta_p_list = np.empty(n_V)
    for j in range(n_V):
        volume = V_list[j]
        pressure = func_pV(volume, B0, BP, V0)
        delta_p_list[j] = abs(pressure - p)
    min_delta_idx = np.argmin(delta_p_list)
    V_guess = V_list[min_delta_idx]
    result = fsolve(f, V_guess)
    V = result[0]
    return V


# end of Birch-Murnaghan equation


def fit_US(entropies, energies, dat_abc):
    guess_abc = dat_abc[0, 1:]
    popt, pcov = curve_fit(func_US, entropies, energies, guess_abc, maxfev=1000000)
    a = popt[0]
    b = popt[1]
    c = popt[2]
    return [a, b, c]


def fit_UV(volumes, energies, dat_eos):
    guess_eos = dat_eos[0, 1:]
    popt, pcov = curve_fit(func_UV, volumes, energies, guess_eos, maxfev=1000000)
    E0 = popt[0]
    B0 = popt[1]
    BP = popt[2]
    V0 = popt[3]
    return [E0, B0, BP, V0]


def at_const_S(S, dat_abc, dat_eos):
    dat = dat_abc
    nd = dat.shape[0]
    volumes = dat[:, 0]
    energies = np.empty(nd)
    for i in range(nd):
        a = dat[i, 1]
        b = dat[i, 2]
        c = dat[i, 3]
        energies[i] = func_US(S, a, b, c)
    [E0, B0, BP, V0] = fit_UV(volumes, energies, dat_eos)
    return [E0, B0, BP, V0]


def at_const_V(V, dat_abc, dat_eos):
    dat = dat_eos
    nd = dat.shape[0]
    entropies = dat[:, 0]
    energies = np.empty(nd)
    for i in range(nd):
        E0 = dat[i, 1]
        B0 = dat[i, 2]
        BP = dat[i, 3]
        V0 = dat[i, 4]
        energies[i] = func_UV(V, E0, B0, BP, V0)
    [a, b, c] = fit_US(entropies, energies, dat_abc)
    return [a, b, c]


def potential_U(S, V, dat_abc, dat_eos):
    [E0, B0, BP, V0] = at_const_S(S, dat_abc, dat_eos)
    p = func_pV(V, B0, BP, V0)
    U1 = func_UV(V, E0, B0, BP, V0)
    [a, b, c] = at_const_V(V, dat_abc, dat_eos)
    T = func_TS(S, a, c)
    U2 = func_US(S, a, b, c)
    U = (U1 + U2) * 0.5
    A = U - T * S
    H = U + p * V
    G = U - T * S + p * V
    return [V, S, p, T, U, A, H, G]


def potential_A(T, V, dat_abc, dat_eos):
    [a, b, c] = at_const_V(V, dat_abc, dat_eos)
    S = func_ST(T, a, c)
    [V, S, p, T, U, A, H, G] = potential_U(S, V, dat_abc, dat_eos)
    return [V, S, p, T, U, A, H, G]


def potential_H(S, p, dat_abc, dat_eos):
    [E0, B0, BP, V0] = at_const_S(S, dat_abc, dat_eos)
    V = func_Vp(p, B0, BP, V0, dat_abc)
    [V, S, p, T, U, A, H, G] = potential_U(S, V, dat_abc, dat_eos)
    return [V, S, p, T, U, A, H, G]


def potential_G(T, p, dat_abc, dat_eos):
    temperature = T
    pressure = p
    nd = dat_eos.shape[0]
    entropies = dat_eos[:, 0]
    enthalpies = np.empty(nd)
    for i in range(nd):
        [V, S, p, T, U, A, H, G] = potential_H(entropies[i], pressure, dat_abc, dat_eos)
        enthalpies[i] = H
    [a, b, c] = fit_US(entropies, enthalpies, dat_abc)
    entropy = func_ST(temperature, a, c)
    [V, S, p, T, U, A, H, G] = potential_H(entropy, pressure, dat_abc, dat_eos)
    return [V, S, p, T, U, A, H, G]



"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2019-2019 by Zhong-Li Liu

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""


import numpy as np
from gibbs_surf import potential_G
from scipy import constants

N_A = constants.N_A


def entropy_calibration(a, b):
    """
    Calibrate the entropy of the liquid phase against the solid phase at reference temperature and pressure.

    Parameters:
    a (float): Reference temperature factor.
    b (float): Reference pressure factor.
    """

    # calibrated state point
    T_ref = float(a) * 1.0e-3   # K ==> K*1e+3
    p_ref = float(b) * 1.0e-1   # GPa ==> Pa*1e+10

    try:
        # solid phase
        dat_abc = np.genfromtxt('solid_dat_abc.txt')
        dat_eos = np.genfromtxt('solid_dat_eos.txt')
        [V, S, p, T, U, F, H, G] = potential_G(T_ref, p_ref, dat_abc, dat_eos)
        G_solid = G

        # liquid phase
        dat_abc = np.genfromtxt('liquid_dat_abc.txt')
        dat_eos = np.genfromtxt('liquid_dat_eos.txt')
        [V, S, p, T, U, F, H, G] = potential_G(T_ref, p_ref, dat_abc, dat_eos)
        V_liquid = V
        S_liquid = S
        p_liquid = p
        T_liquid = T
        U_liquid = U

        S_before_shift = S_liquid
        S_after_shift = (U_liquid + p_liquid * V_liquid - G_solid) / T_liquid
        delta_sx = S_after_shift - S_before_shift

        # Write result to file
        with open('liquid_delta_S.txt', 'w') as f:
            f.write(f"{delta_sx:.6f}")

    except Exception as e:
        print(f"An error occurred: {e}")



"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2019-2019 by Zhong-Li Liu

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""

import numpy as np
from scipy import constants

# Constants for unit conversion
kBar_tO_internal_p = 1e-2  # kBar ==> 1e+10 Pa
eV_tO_internal_U = constants.e * 1e+19  # eV/atom ==> 1e-19 J/atom
K_tO_internal_T = 1e-3  # K ==> 1.0e+3 K
volume_to_internal_V = 1e-1  # A^3/atom ==> 1e-29 m^3/atom
precision = 6  # Number of decimal places to retain


def read_data(input_file):
    """Read data from the input file."""
    dat = np.genfromtxt(input_file)
    return dat


def write_mid_temperature(dat):
    """Calculate and write the middle temperature to mid_T.txt."""
    # mid_T = np.mean(dat[:, 3]) * K_tO_internal_T
    max_T = max(dat[:, 3])
    min_T = min(dat[:, 3])
    mid_T = (max_T + min_T) * 0.5 * 1e-3

    with open('mid_T.txt', 'w') as middle_T:
        # middle_T.write(f"{mid_T:.{precision}f}\n")
        middle_T.write(f"{mid_T:.3f}\n")

def write_volumes(unique_volumes):
    """Write unique volumes to V.txt."""

    with open('V.txt', 'w') as all_volume:
        all_volume.write("# V(*1e-29 m^3/atom)\n")
        for v in unique_volumes:
            all_volume.write(f"{v * volume_to_internal_V:.{precision}f}\n")


def write_single_volumes(dat, unique_volumes):
    """Write pressure, energy, and temperature for each unique volume."""
    for j, v in enumerate(unique_volumes, start=1):

        with open(f'V{j}.txt', 'w') as single_volume:
            single_volume.write("# T(*1e+3 K), U(*1e-19 J/atom), p(*1e+10 Pa)\n")
            for i in range(dat.shape[0]):
                if v == dat[i, 0]:
                    this_pressure = dat[i, 1] * kBar_tO_internal_p
                    this_energy = dat[i, 2] * eV_tO_internal_U
                    this_temperature = dat[i, 3] * K_tO_internal_T
                    single_volume.write(
                        f"{this_temperature:.{precision}f} "
                        f"{this_energy:.{precision}f} "
                        f"{this_pressure:.{precision}f}\n")


def surface_01():
    """Read input files of two phase and convert units."""
    input_file = 'input.txt'
    dat = read_data(input_file)

    write_mid_temperature(dat)

    unique_volumes = np.unique(dat[:, 0])[::-1]
    write_volumes(unique_volumes)
    write_single_volumes(dat, unique_volumes)



"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2019-2019 by Zhong-Li Liu

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""

# find intersection of G-T curves at constant P

import numpy as np
from scipy.optimize import fsolve
import matplotlib.pyplot as plt
from scipy import constants
from gibbs_surf import potential_G
import json
import sys

eV = constants.eV


def get_cross_point(level_str, cp, x_label, y_label, y_label_vasp, n):
    """
    Find the intersection point of Gibbs free energy curves for solid and liquid phases.

    Parameters:
    level_str (str): The pressure level as a string.
    cp (file object): The file object to write the melting point.
    x_label (str): Label for the x-axis of the plot.
    y_label (str): Label for the y-axis of the plot.
    n (str): The filename prefix for the plot.

    Returns:
    float: The melting temperature.
    """
    try:
        dat_solid = np.genfromtxt(f'solid_pressure_{level_str}.txt')
        dat_liquid = np.genfromtxt(f'liquid_pressure_{level_str}.txt')
        order = 3

        # Solid phase data
        x_solid = dat_solid[:, 0]
        y_solid = dat_solid[:, 1]
        Ga = np.polyfit(x_solid, y_solid, order)
        plt.clf()
        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.plot(x_solid, y_solid, 'r+', label='Solid phase')

        # Liquid phase data
        x_liquid = dat_liquid[:, 0]
        y_liquid = dat_liquid[:, 1]
        Gb = np.polyfit(x_liquid, y_liquid, order)
        plt.plot(x_liquid, y_liquid, 'gx', label='Liquid phase')

        # Look for intersection
        lower_bound = max(min(x_solid), min(x_liquid))
        upper_bound = min(max(x_solid), max(x_liquid))
        xs = np.linspace(lower_bound, upper_bound, 100)
        ys_solid = np.polyval(Ga, xs)
        ys_liquid = np.polyval(Gb, xs)
        guess = 0.5 * (lower_bound + upper_bound)
        f = np.poly1d(Ga - Gb)
        root = fsolve(f, guess)
        temperature = root[0]
        gibbsenergy = 0.5 * (np.polyval(Ga, temperature) + np.polyval(Gb, temperature))
        cp.write(f"{level_str} {temperature:.3f}\n")

        plt.plot(temperature, gibbsenergy, 'bo', label='Melting point', zorder=5)
        plt.plot(xs, ys_solid, 'r-', label='Solid fit')
        plt.plot(xs, ys_liquid, 'g-', label='Liquid fit')
        plt.tick_params(axis='both', which='both', direction='in', top=True, right=True)
        plt.minorticks_on()
        plt.legend(frameon=False)
        plt.savefig(f'{n}_pressure_{level_str}_internal.pdf', format='pdf')  # save image with internal units

        # convert units from internal to vasp type
        y_solid_ev = y_solid / (eV * 1.0e+19)
        y_liquid_ev = y_liquid / (eV * 1.0e+19)
        ys_solid_ev = ys_solid / (eV * 1.0e+19)
        ys_liquid_ev = ys_liquid / (eV * 1.0e+19)
        gibbsenergy_ev = gibbsenergy / (eV * 1.0e+19)

        plt.clf()
        plt.xlabel(x_label)
        plt.ylabel(y_label_vasp)
        plt.plot(x_solid, y_solid_ev, 'r+', label='Solid phase')
        plt.plot(x_liquid, y_liquid_ev, 'gx', label='Liquid phase')
        plt.plot(temperature, gibbsenergy_ev, 'bo', label='Melting point', zorder=5)
        plt.plot(xs, ys_solid_ev, 'r-', label='Solid fit')
        plt.plot(xs, ys_liquid_ev, 'g-', label='Liquid fit')
        plt.tick_params(axis='both', which='both', direction='in', top=True, right=True)
        plt.minorticks_on()
        plt.legend(frameon=False)
        plt.savefig(f'{n}_pressure_{level_str}_vasp.pdf', format='pdf')

        return temperature
    except Exception as e:
        print(f"An error occurred while finding the cross point: {e}", file=sys.stderr)
        return None


def get_melting_data(p, n):
    """
    Generate and save melting data for given pressure.

    Parameters:
    p (float): The pressure level.
    n (str): The filename prefix for the input JSON file.
    """
    try:
        level = float(p) * 1.0e-1  # GPa ==> Pa*1e+10
        level_str_c = f"{float(p):.3f}"
        x_label = r'Temperature ($\times 10^{3}$ K)'
        y_label = r'Gibbs free energy ($\times 10^{-19}$ J/atom)'
        y_label_vasp = r'Gibbs free energy (eV/atom)'

        with open(f'cross_point_{level_str_c}.txt', 'w') as cp:
            cp.write("# P(* GPa), T(*1e+3 K)\n")
            get_cross_point(level_str_c, cp, x_label, y_label, y_label_vasp, n)
    except Exception as e:
        print(f"An error occurred while getting the melting data: {e}", file=sys.stderr)




"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2019-2019 by Zhong-Li Liu

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""

import numpy as np
from gibbs_surf import *
import sys


def generate_derived_surface(phase='solid', nV='nV', nT='nT'):
    """
    Generate derived surface data for a given phase.

    Parameters:
    phase (str): The phase for which to generate the surface data ('solid' or 'liquid').
    nV (str): Number of volume data points.
    nT (str): Number of temperature data points.
    """

    try:
        # Load data files
        dat_abc = np.genfromtxt(f'{phase}_dat_abc.txt')
        dat_eos = np.genfromtxt(f'{phase}_dat_eos.txt')
        data = np.genfromtxt(f'{phase}_input.txt')

        # Extract temperature data and compute min and max values
        temperatures = data[:, 3]
        T_max = np.max(temperatures) * 1e-3
        T_min = np.min(temperatures) * 1e-3

        # Extract volume data and compute min and max values
        volumes = dat_abc[:, 0]
        V_min = min(volumes)
        V_max = max(volumes)

        # Open output file
        with open(f'data_{phase}.txt', 'w') as f:
            # Generate volume and temperature data points
            V_data = np.linspace(V_min, V_max, int(nV))
            T_data = np.linspace(float(T_min), float(T_max), int(nT))

            # Calculate properties for each combination of V and T
            for V in V_data:
                for T in T_data:
                    V, S, P, T, E, A, H, G = potential_A(T, V, dat_abc, dat_eos)
                    f.write(f"{V:10.6f} {S:10.6f} {P:10.6f} {T:10.6f} {E:10.6f} {A:10.6f} {H:10.6f} {G:10.6f}\n")

    except Exception as e:
        print(f"An error occurred: {e}", file=sys.stderr)



import json
import numpy as np


def two_phase(phase):
    p = str(phase)
    # phase_dict = {p: {}}
    phase_dict = {}
    with open('%s' % p + '_dat_abc.txt', 'r') as f1:
        abc_lines = f1.readlines()

    abc_list = []
    for line in abc_lines:
        dat1 = [float(value) for value in line.split()]
        abc_list.append(dat1)

    with open('%s' % p + '_dat_eos.txt', 'r') as f2:
        eos_lines = f2.readlines()

    eos_list = []
    for line in eos_lines:
        dat2 = [float(value) for value in line.split()]
        eos_list.append(dat2)

    with open('data_' + '%s' % p + '.txt', 'r') as f3:
        s_eos_lines = f3.readlines()

    dat_list = []
    for line in s_eos_lines:
        dat3 = [float(value) for value in line.split()]
        dat_list.append(dat3)

    phase_dict["abc"] = abc_list
    phase_dict['eos'] = eos_list
    phase_dict['dat'] = dat_list
    return phase_dict


def save_json(substance_name, phase1, phase2):
    name = str(substance_name)
    dat = {phase1: two_phase(phase1), phase2: two_phase(phase2)}
    with open('%s' % name + '.json', 'w') as json_file:
        json.dump(dat, json_file)



"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2019-2019 by Zhong-Li Liu

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit, leastsq


def eos(V, B0, BP, V0):
    """3rd-order Birch-Murnaghan Equation of State (EoS)."""
    eta = (V0 / V) ** (1. / 3.)
    return 1.5 * B0 * (eta ** 7 - eta ** 5) * (1 + 3. / 4. * (BP - 4) * (eta ** 2 - 1))


def A_without_constant(V, B0, BP, V0):
    """Helmholtz free energy by integration of EoS (without constant A0)."""
    eta = (V0 / V) ** (1. / 3.)
    return 9. * V0 * B0 / 16. * ((eta ** 2 - 1) ** 3 * BP + (eta ** 2 - 1) ** 2 * (6 - 4 * eta ** 2))


def func_ST(T, a, c):
    """Calculate entropy S as a function of T."""
    return a * np.log(a * T / c)


def func_TS(S, a, c):
    """Calculate temperature T as a function of S."""
    return c / a * np.exp(S / a)


def func_UV(V, E0, B0, BP, V0):
    """Calculate internal energy U as a function of V."""
    eta = (V0 / V) ** (1. / 3.)
    return E0 + 9. * B0 * V0 / 16. * ((eta ** 2 - 1) ** 3 * BP + (eta ** 2 - 1) ** 2 * (6 - 4 * eta ** 2))


def bm3(parameters, V):
    """Birch-Murnaghan 3rd-order EoS for optimization."""
    E0, B0, BP, V0 = parameters
    eta = (V0 / V) ** (1. / 3.)
    return E0 + 9. * B0 * V0 / 16. * ((eta ** 2 - 1) ** 3 * BP + (eta ** 2 - 1) ** 2 * (6 - 4 * eta ** 2))


def objective(pars, y, x):
    """Objective function for least squares optimization."""
    return y - bm3(pars, x)


def func_US(S, a, b, c):
    """Calculate internal energy U as a function of entropy S."""
    return c * np.exp(S / a) + b


def parabola(x, a, b, c):
    """Parabolic function for fitting."""
    return a + b * x + c * x ** 2


def read_data(file_path):
    """Read data from a file."""
    return np.genfromtxt(file_path)


def write_data(file_path, data):
    """Write data to a file."""
    try:
        with open(file_path, 'w') as f:
            for line in data:
                f.write(line + '\n')
    except IOError as e:
        print(f"Error writing to file {file_path}: {e}")


def surface_03():
    """Main function for Gibbs Thermodynamic Surface analysis."""

    # whether output full information and intermediate graphics to debug
    debug = False

    # load data
    V = read_data('V.txt')
    if V is None:
        return

    nV = len(V)
    dat = read_data('abde.txt')
    if dat is None:
        return

    a, b, d, e = dat[:, 0], dat[:, 1], dat[:, 2], dat[:, 3]

    # set properties at the reference state
    refS = read_data('refS.txt')
    ref_index = int(read_data('ref_index.txt'))
    refV = V[ref_index]
    write_data('refV.txt', [f"{refV:.6f}"])

    refT = read_data('refT.txt')
    refU = a[ref_index] * refT + b[ref_index]

    # get V-p data at the reference T
    p = d * refT + e

    # fit V-p to Equation of State (EoS)
    popt, pcov = curve_fit(eos, V, p, maxfev=1000000)
    B0, BP, V0 = popt

    if debug:
        print('EoS parameters for A_without_constant:')
        print(f'B0 = {B0:.6f}, BP = {BP:.6f}, V0 = {V0:.6f}')
        print(p)

    # derive preliminary coefficient c
    file_out = 'preliminary_c.txt'
    preliminary_c_data = []
    for i in range(nV):
        thisV = V[i]
        thisU = a[i] * refT + b[i]
        thisA = A_without_constant(thisV, B0, BP, V0)
        refA = A_without_constant(refV, B0, BP, V0)
        thisS = ((thisU - refU + refT * refS) - (thisA - refA)) / refT
        c = a[i] * refT / np.exp(thisS / a[i])
        preliminary_c_data.append(f"{c:.6f}")
    write_data(file_out, preliminary_c_data)

    # generate coefficients of U(V) at constant S, i.e., dat_eos.txt
    c = read_data('preliminary_c.txt')
    if c is None:
        return

    S_all = []
    for i in range(nV):
        file_tup = f'V{i + 1}.txt'
        dat = read_data(file_tup)
        if dat is None:
            continue

        T = dat[:, 0]
        for thisT in T:
            thisS = func_ST(thisT, a[i], c[i])
            S_all.append(thisS)

    minS, maxS = min(S_all), max(S_all)

    if debug:
        print(f"Entropy range: minimum = {minS:.6f}, maximum = {maxS:.6f}")
        for i in range(nV):
            mintemp = func_TS(minS, a[i], c[i])
            maxtemp = func_TS(maxS, a[i], c[i])
            print(f"V = {V[i]:.6f}, minimum T = {mintemp:.6f}, maximum T = {maxtemp:.6f}")

    file_out = 'dat_eos.txt'
    dat_eos_data = []
    nS = 10  # number of entropies to sample
    S = np.linspace(minS, maxS, nS)

    fig = plt.figure()
    plt.subplot(111)
    for thisS in S:
        U = np.array([func_US(thisS, a[j], b[j], c[j]) for j in range(nV)])

        # plot figure
        if debug:
            plt.xlabel('V (1.0e-29 m^3/atom)')
            plt.ylabel('U (1.0e-19 J/atom)')
            plt.title(f'Constant S = {thisS:.3f} * 1.0e-22 J/K/atom')
            plt.plot(V, U, 'o')

        # estimate the initial guess
        p0 = [min(U), 1, 1]
        popt, pcov = curve_fit(parabola, V, U, p0, maxfev=1000000)
        a0, b0, c0 = popt
        parabola_vmin = -b0 / (2 * c0)
        E00 = parabola(parabola_vmin, a0, b0, c0)
        B00 = 2 * c0 * parabola_vmin
        BP0 = 4
        initial_guess = [E00, B00, BP0, parabola_vmin]
        guess = initial_guess if 'guess' not in locals() else guess
        plsq = leastsq(objective, guess, args=(U, V), maxfev=1000000)
        popt = plsq[0]
        E0, B0, BP, V0 = popt
        guess = [E0, B0, BP, V0]
        dat_eos_data.append(f"{thisS:.6f} {E0:.6f} {B0:.6f} {BP:.6f} {V0:.6f}")

        if debug:
            print(f"{thisS:.6f} {E0:.6f} {B0:.6f} {BP:.6f} {V0:.6f}")
            nk = 50
            V_s = np.linspace(min(V), max(V), nk)
            U_s = np.array([func_UV(V_s[k], E0, B0, BP, V0) for k in range(nk)])
            plt.plot(V_s, U_s, 'k-')
            plt.show()

    write_data(file_out, dat_eos_data)

    # generate coefficients of U(S) at constant V again, i.e., dat_abc.txt
    dat_abde = read_data('abde.txt')
    if dat_abde is None:
        return

    dat_c = read_data('preliminary_c.txt')
    if dat_c is None:
        return

    guess = [dat_abde[0, 0], dat_abde[0, 1], dat_c[0]]

    file_out = 'dat_abc.txt'
    dat_abc_data = []
    dat_eos = read_data('dat_eos.txt')
    if dat_eos is None:
        return

    S = dat_eos[:, 0]
    nS = dat_eos.shape[0]
    for i in range(nV):
        thisV = V[i]
        U = np.array([func_UV(thisV, dat_eos[j, 1], dat_eos[j, 2], dat_eos[j, 3], dat_eos[j, 4]) for j in range(nS)])
        popt, pcov = curve_fit(func_US, S, U, guess, maxfev=1000000)
        a, b, c = popt
        dat_abc_data.append(f"{thisV:.6f} {a:.6f} {b:.6f} {c:.6f}")
    write_data(file_out, dat_abc_data)






"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2019-2019 by Zhong-Li Liu

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""

import json
import numpy as np
from gibbs_surf import potential_G
from scipy import constants
import sys


def print_internal(data, name):
    """
    Print internal thermodynamic properties for solid and liquid phases.

    Parameters:
    data (str): Path to the data file containing melting pressure and temperature.
    name (str): Name of the material, used to load the corresponding JSON file.
    """

    try:
        with open(data, 'r') as file:
            lines = file.readlines()

        # Extract the melting pressure and temperature
        temp_pressure_line = lines[1].strip().split()
        pressure_melting = float(temp_pressure_line[0])
        temp_melting = float(temp_pressure_line[1])  # Convert to K

        properties = ['G', 'A', 'H', 'U', 'S', 'V']
        units = ['1.0e-19 J/atom', '1.0e-19 J/atom', '1.0e-19 J/atom',
                 '1.0e-19 J/atom', r'1.0e-22 J/K/atom', '1.0e-29 m**3/atom']

        print(f"melting_pressure {'=':>4} {pressure_melting:.3f} GPa")
        print(f"melting_temperature {'=':>1} {temp_melting * 1.0e+3:.1f} K")
        # print(f"##{properties[0]}:{units[0]}, {properties[1]}:{units[1]}, {properties[2]}:{units[2]}##\n"
        #       f"##{properties[3]}:{units[3]}, {properties[4]}:{units[4]}, {properties[5]}:{units[5]}##")
        print(f"{'####':^6} {'#solid#':^9}{'#liquid#':^10}{'#units#':^16}")

        # start to print thermodynamic potential
        with open('%s' % str(name) + '.json', 'r') as jsonfile:
            material = json.load(jsonfile)
        [V, S, p, T, U, A, H, G] = potential_G(temp_melting, pressure_melting * 1.0e-1, np.array(material['solid']['abc']),
                                               np.array(material['solid']['eos']))
        G_sol = G
        A_sol = A
        H_sol = H
        U_sol = U
        S_sol = S
        V_sol = V

        [V, S, p, T, U, A, H, G] = potential_G(temp_melting, pressure_melting * 1.0e-1, np.array(material['liquid']['abc']),
                                               np.array(material['liquid']['eos']))
        G_liq = G
        A_liq = A
        H_liq = H
        U_liq = U
        S_liq = S
        V_liq = V

        print(f"{properties[0]:^5}{G_sol:>9.3f}{G_liq:>10.3f}{units[0]:^20}")
        print(f"{properties[1]:^5}{A_sol:>9.3f}{A_liq:>10.3f}{units[1]:^20}")
        print(f"{properties[2]:^5}{H_sol:>9.3f}{H_liq:>10.3f}{units[2]:^20}")
        print(f"{properties[3]:^5}{U_sol:>9.3f}{U_liq:>10.3f}{units[3]:^20}")
        print(f"{properties[4]:^5}{S_sol:>9.3f}{S_liq:>10.3f}{units[4]:^20}")
        print(f"{properties[5]:^5}{V_sol:>9.3f}{V_liq:>10.3f}{units[5]:^20}")

    except Exception as e:
        print(f"An error occurred: {e}", file=sys.stderr)


# convert units to VASP standard
def print_vasp_type(data, name):
    """
    Print thermodynamic properties in VASP units for solid and liquid phases.

    Parameters:
    data (str): Path to the data file containing melting pressure and temperature.
    name (str): Name of the material, used to load the corresponding JSON file.
    """

    eV = constants.e

    try:
        with open(data, 'r') as file:
            lines = file.readlines()

        # Extract the melting pressure and temperature
        temp_pressure_line = lines[1].strip().split()
        pressure_melting = float(temp_pressure_line[0])
        temp_melting = float(temp_pressure_line[1])

        properties = ['G', 'A', 'H', 'U', 'S', 'V']
        units = ['eV/atom', 'eV/atom', 'eV/atom',
                 'eV/atom', 'eV/K/atom', 'A**3/atom']

        print(f"melting_pressure {'=':>4} {pressure_melting:.3f} GPa")
        print(f"melting_temperature {'=':>1} {temp_melting * 1.0e+3:.1f} K")
        # print(f"##{properties[0]}:{units[0]}, {properties[1]}:{units[1]}, {properties[2]}:{units[2]}##\n"
        #       f"##{properties[3]}:{units[3]}, {properties[4]}:{units[4]}, {properties[5]}:{units[5]}##")
        print(f"{'####':^6} {'#solid#':^9}{'#liquid#':^10}{'#units#':^16}")

        # start to print thermodynamic potential
        with open('%s' % str(name) + '.json', 'r') as jsonfile:
            material = json.load(jsonfile)
        [V, S, p, T, U, A, H, G] = potential_G(temp_melting, pressure_melting * 1.0e-1, np.array(material['solid']['abc']),
                                               np.array(material['solid']['eos']))
        G_sol = float(G) / (eV * 1.0e+19)
        A_sol = float(A) / (eV * 1.0e+19)
        H_sol = float(H) / (eV * 1.0e+19)
        U_sol = float(U) / (eV * 1.0e+19)
        S_sol = float(S) * 1.0e-3 / (eV * 1.0e+19)
        V_sol = float(V) * 1.0e+1

        [V, S, p, T, U, A, H, G] = potential_G(temp_melting, pressure_melting * 1.0e-1, np.array(material['liquid']['abc']),
                                               np.array(material['liquid']['eos']))
        G_liq = float(G) / (eV * 1.0e+19)
        A_liq = float(A) / (eV * 1.0e+19)
        H_liq = float(H) / (eV * 1.0e+19)
        U_liq = float(U) / (eV * 1.0e+19)
        S_liq = float(S) * 1.0e-3 / (eV * 1.0e+19)
        V_liq = float(V) * 1.0e+1

        # Print the properties
        print(f"{properties[0]:^5}{G_sol:>9.3f}{G_liq:>10.3f}{units[0]:^20}")
        print(f"{properties[1]:^5}{A_sol:>9.3f}{A_liq:>10.3f}{units[1]:^20}")
        print(f"{properties[2]:^5}{H_sol:>9.3f}{H_liq:>10.3f}{units[2]:^20}")
        print(f"{properties[3]:^5}{U_sol:>9.3f}{U_liq:>10.3f}{units[3]:^20}")
        print(f"{properties[4]:^5}{S_sol:>9.6f}{S_liq:>10.6f}{units[4]:^20}")
        print(f"{properties[5]:^5}{V_sol:>9.3f}{V_liq:>10.3f}{units[5]:^20}")

    except Exception as e:
        print(f"An error occurred: {e}", file=sys.stderr)



"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2024-2024 by Kun Yin and Xuan Zhao

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""

from setuptools import setup, find_packages
import os

with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="GibbsTS",
    version="1.0.3",
    description="GTS: An automated toolkit for building Gibbs thermodynamic surface with application to obtain "
                "high-pressure melting data.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Kun Yin, Xuan Zhao",
    author_email="yinkun@cdut.edu.cn",
    license="GNU General Public License v3.0",
    packages=find_packages(where='lib'),
    package_dir={'': 'lib'},
    py_modules=["gibbs_thermo_surface_01", "gibbs_thermo_surface_02", "gibbs_thermo_surface_03",
                "gibbs_thermo_surface_04", "gibbs_thermo_surface_05", "gibbs_thermo_surface_06",
                "gibbs_thermo_surface_07", "gibbs_thermo_surface_08", "gibbs_surf", "save_json"],
    scripts=["lib/GTS"],
    python_requires='>=3.11',
    install_requires=[
        'numpy>=1.24.3',
        'scipy>=1.10.1',
        'matplotlib>=3.7.1',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Physics',
    ],

)




"""
  GTS -- Gibbs Thermodynamic Surface: an automated toolkit to obtain high-pressure melting data

  Copyright (C) 2019-2019 by Zhong-Li Liu

  This program is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software Foundation
  version 3 of the License.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  E-mail: yinkun@cdut.edu.cn
"""

import numpy as np


def fit_linear_functions():
    """Fit linear functions U(T)=a*T+b and p(T)=d*T+e for data in V*.txt files."""

    volumes = np.genfromtxt('V.txt')  # Read the volumes from V.txt

    file_out = 'abde.txt'

    with open(file_out, 'w') as f:
        for i in range(len(volumes)):
            file_tup = f'V{i + 1}.txt'

            dat = np.genfromtxt(file_tup)

            T = dat[:, 0]
            U = dat[:, 1]
            p = dat[:, 2]
            [a, b] = np.polyfit(T, U, 1)
            [d, e] = np.polyfit(T, p, 1)
            f.write(f"{a:.6f} {b:.6f} {d:.6f} {e:.6f}\n")
